# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for TFX Firebase Publisher Custom Component."""

import tensorflow as tf
from tfx.types import standard_artifacts

from tfx_addons.firebase_publisher.component import FirebasePublisher


class FirebasePublisherTest(tf.test.TestCase):
  def testConstruct(self):
    firebase_publisher = FirebasePublisher(display_name="test_display_name",
                                           storage_bucket="storage_bucket")

    self.assertEqual(standard_artifacts.PushedModel.TYPE_NAME,
                     firebase_publisher.outputs['pushed_model'].type_name)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""PandasTransform TFX component.  Feature engineering in TFX using Pandas Dataframes."""

import os

import apache_beam as beam
import pandas as pd
import tensorflow as tf
import tensorflow_data_validation as tfdv
from absl import logging
from packaging.version import Version, parse
from tensorflow_transform.tf_metadata import schema_utils
from tfx import v1 as tfx
from tfx.components.util import tfxio_utils
from tfx.dsl.component.experimental.decorators import component
from tfx.types import artifact_utils
from tfx.types.standard_artifacts import Examples, ExampleStatistics, Schema
from tfx.utils import import_utils, io_utils

if parse(tfx.__version__) >= Version('1.8.0'):
  from tfx.dsl.component.experimental.annotations import BeamComponentParameter
  from tfx.types import system_executions

_TELEMETRY_DESCRIPTORS = ['PandasTransform']


class Arrow2PandasTypes(beam.DoFn):
  """Beam DoFn that converts from the Arrow types that are generated by TFXIO to Pandas types."""
  def __init__(self):
    pass

  # TODO: This does not correctly process features with valence > 1
  def process(self, element, schema=None):
    """Processes each dataframe to change types based on the schema.
    Args:
      element: a Pandas dataframe, converted from an Arrow RecordBatch.
      schema: a Python dict of a schema proto created by SchemaGen.
    Returns:
      Updated dataframe
    """
    if schema is None:
      raise ValueError('Arrow2PandasTypes: Schema is required')

    for key, feature in element.items():
      for idx, _ in enumerate(feature):
        if feature[idx].size > 0:
          feature[idx] = feature[idx][0]
          if schema[key] == 'string':
            feature[idx] = feature[idx].decode('utf8')
        else:
          feature[idx] = None
      feature = feature.astype(schema[key], copy=False)

    yield element


class GetExamples(beam.DoFn):
  """Beam DoFn that created TF.train.Examples from the rows of a dataframe."""
  def __init__(self):
    pass

  def DictToExample(self, row, ptypes):
    """Creates TF.train.Examples containing tf.train.Features from a dataframe
    Args:
      row: A dict containing a row from a dataframe
      ptypes: A dict of the Pandas types for this dataframe
    Returns:
      The corresponding tf.train.Example
    """
    feature = {}
    for key, val in row.items():
      val = [] if pd.isna(val) else [val]
      if ptypes[key] == 'Int64':
        feature[key] = tf.train.Feature(int64_list=tf.train.Int64List(
            value=val))
      elif ptypes[key] in ['float32', 'Float64']:
        feature[key] = tf.train.Feature(float_list=tf.train.FloatList(
            value=val))
      elif ptypes[key] in ['string', 'object']:
        if val != []:
          val = [val[0].encode('utf8')]
        feature[key] = tf.train.Feature(bytes_list=tf.train.BytesList(
            value=val))
      else:
        raise ValueError('DictToExample: Type {} was unhandled'.format(
            ptypes[key]))
    return tf.train.Example(features=tf.train.Features(feature=feature))

  def process(self, df):
    """The Beam DoFn process function.  Yields serialized tf.train.Examples
      Args:
        df: A Pandas dataframe
      Returns:
        Yields serialized tf.train.Examples
    """
    ptypes = df.dtypes.apply(lambda x: x.name).to_dict()
    for row_dict in df.to_dict(orient='records'):
      example = self.DictToExample(row_dict, ptypes)
      yield example.SerializeToString()


if parse(tfx.__version__) >= Version('1.8.0'):

  @component(component_annotation=system_executions.Transform, use_beam=True)
  def PandasTransform(
      transformed_examples: tfx.dsl.components.OutputArtifact[Examples],
      examples: tfx.dsl.components.InputArtifact[Examples] = None,
      schema: tfx.dsl.components.InputArtifact[Schema] = None,
      statistics: tfx.dsl.components.InputArtifact[ExampleStatistics] = None,
      module_file: tfx.dsl.components.Parameter[str] = None,
      beam_pipeline_args: tfx.dsl.components.Parameter[str] = None,
      beam_pipeline: BeamComponentParameter[beam.Pipeline] = None,
  ) -> None:
    """This docstring will be replaced by the shared docstring below"""
    if beam_pipeline_args is not None:
      this_beam_pipeline = beam.Pipeline(argv=beam_pipeline_args.split(' '))
    else:
      this_beam_pipeline = beam_pipeline

    DoPandasTransform(examples=examples,
                      schema=schema,
                      statistics=statistics,
                      transformed_examples=transformed_examples,
                      module_file=module_file,
                      beam_pipeline=this_beam_pipeline)
else:
  logging.info('TFX < 1.8.0')
  logging.info('Beam custom component not supported in this version of TFX.')
  logging.info('You might consider an upgrade.')

  @component
  def PandasTransform(
      transformed_examples: tfx.dsl.components.OutputArtifact[Examples],
      examples: tfx.dsl.components.InputArtifact[Examples] = None,
      schema: tfx.dsl.components.InputArtifact[Schema] = None,
      statistics: tfx.dsl.components.InputArtifact[ExampleStatistics] = None,
      module_file: tfx.dsl.components.Parameter[str] = None,
      beam_pipeline_args: tfx.dsl.components.Parameter[str] = None) -> None:
    """This docstring will be replaced by the shared docstring below"""
    if beam_pipeline_args is not None:
      this_beam_pipeline = beam.Pipeline(argv=beam_pipeline_args.split(' '))
    else:
      this_beam_pipeline = beam.Pipeline()

    DoPandasTransform(examples=examples,
                      schema=schema,
                      statistics=statistics,
                      transformed_examples=transformed_examples,
                      module_file=module_file,
                      beam_pipeline=this_beam_pipeline)


PandasTransform.__doc__ = """The PandasTransform TFX component.
  PandasTransform enables users to perform feature engineering in dataframes, using
  either Pandas, Modin, Numpy, Scikit, or another library which supports dataframes.
  **Important Note:** Because processing is distributed by Beam across a cluster of
  compute nodes, each invocation of PandasTransform will recieve a subset of the data.
  That means that **operations which require full passes over the dataset are not supported.**
  If you require full passes over the dataset, you are encouraged to use TensorFlow
  Transform instead.  However, since PandasTransform is designed to receive a dict of the
  summary statistics which are created by StatisticsGen (or other tooling) then you
  can often use those values to avoid making a full pass over the data.
  **Important Note:** Unlike TensorFlow Transform, PandasTransform does not create a
  preprocessing TensorFlow graph to be prepended to your trained model.  That means
  that you **MUST** perform the equivalent preprocessing in your serving client or
  a similar location in your system.

  Args:
    examples: A TFX input channel containing a dataset artifact
    schema: A TFX input channel containing a schema artifact
    statistics: A TFX input channel containing a statistics artifact
    transformed_examples: A TFX output channel which will be used to output the resulting
    dataset artifact
    module_file: A component parameter containing a file path to a Python file which
    contains the user code, in a function named 'preprocessing_fn'.
    beam_pipeline_args: A string with the argv options for creating a Beam pipeline.
    Note that this is a string, not a list.  It will be split on spaces to create
    a list. If running TFX >= 1.8.0, if beam_pipeline_args are specified they will
    override the pipeline beam args.


  Returns:
    The resulting dataset artifact after processing by the user code.

  Raises:
    ImportError - When the module file is not found.

  Example:
    from tfx_addons.pandas_transform import PandasTransform

    module_file = 'my_module_file.py'
    %%writefile {module_file} # Assuming a Jupyter notebook, otherwise a file
        import numpy as np

        def preprocessing_fn(df, schema, statistics):
        def zscore(x, stats=None):
            if x.isnull()[0] or stats[x.name]['std_dev'] == 0.0:
            return x
            else:
            y = (x - stats[x.name]['mean']) / stats[x.name]['std_dev']
            return y.astype(x.dtype) # to maintain type consistency
            
        numeric_cols = df.select_dtypes(include=[np.number]).columns
        return df[numeric_cols].apply(zscore, stats=statistics)

    _beam_pipeline_args = [
      '--direct_running_mode=in_memory',
      '--direct_num_workers=1']

    panda = PandasTransform(
        examples=example_gen.outputs['examples'],
        schema=schema_gen.outputs['schema'],
        statistics=statistics_gen.outputs['statistics'],
        module_file = os.path.abspath(module_file),
        beam_pipeline_args=' '.join(_beam_pipeline_args)
        )
    context.run(panda, enable_cache=False) # Assuming interactive context
  """


def DoPandasTransform(
    transformed_examples: tfx.dsl.components.OutputArtifact[Examples],
    examples: tfx.dsl.components.InputArtifact[Examples],
    schema: tfx.dsl.components.InputArtifact[Schema],
    statistics: tfx.dsl.components.InputArtifact[ExampleStatistics],
    module_file: tfx.dsl.components.Parameter[str],
    beam_pipeline: beam.pipeline.Pipeline):
  """The function where the actual transforms are done, for both signatures."""
  local_module_file = io_utils.ensure_local(module_file)
  if not os.path.exists(local_module_file):
    raise ImportError(
        f'DoPandasTransform: Module file not found: {module_file}')
  elif examples is None:
    raise ValueError('DoPandasTransform: examples cannot be None')
  elif schema is None:
    raise ValueError('DoPandasTransform: schema cannot be None')
  elif statistics is None:
    raise ValueError('DoPandasTransform: statistics cannot be None')
  elif beam_pipeline is None:
    raise ValueError('DoPandasTransform: beam_pipeline cannot be None')

  def GetFeatureStats(view, feature_name):
    return view.get_feature(tfdv.types.FeaturePath([feature_name])).proto()

  def GetRawFeatureSpec(schema):
    return schema_utils.schema_as_feature_spec(schema).feature_spec

  def WrapUserCode(df, schema=None, statistics=None):
    yield user_code(df, schema, statistics)

  # Get input splits
  input_examples = artifact_utils.get_single_instance([examples])
  examples_split_names = artifact_utils.decode_split_names(
      input_examples.split_names)

  # Init output splits
  transformed_examples = artifact_utils.get_single_instance(
      [transformed_examples])
  transformed_examples.split_names = artifact_utils.encode_split_names(
      examples_split_names)
  for split in examples_split_names:
    os.mkdir(os.path.join(transformed_examples.uri, 'Split-{}'.format(split)))

  # Get the preprocessing_fn from the module file
  user_code = import_utils.import_func_from_source(module_file,
                                                   'preprocessing_fn')

  # Get statistics
  stats_artifact = artifact_utils.get_single_instance([statistics])
  stats_uri = io_utils.get_only_uri_in_dir(
      artifact_utils.get_split_uri([stats_artifact], 'train'))
  stats = tfdv.load_stats_binary(stats_uri)
  stats_view = tfdv.utils.stats_util.DatasetListView(stats).get_default_slice()

  # Get the schema
  schema_file = io_utils.get_only_uri_in_dir(
      artifact_utils.get_single_uri([schema]))
  schema_reader = io_utils.SchemaReader()
  tf_schema = schema_reader.read(schema_file)
  fspec = GetRawFeatureSpec(tf_schema)

  # Convert the statistics and schema to dictionaries
  schema_dict = {}
  schema_dict['domains'] = {}
  stats_dict = {}
  for key, _ in fspec.items():
    if fspec[key].dtype.name == 'int64':
      dtype = 'Int64'
    else:
      dtype = fspec[key].dtype.name
    schema_dict[key] = dtype
    feature = tfdv.utils.schema_util.get_feature(tf_schema, key)
    if tfdv.utils.schema_util.is_categorical_feature(feature):
      schema_dict['domains'][key] = list(
          tfdv.utils.schema_util.get_domain(tf_schema, key).value)

    if dtype in ['Int64', 'float32']:
      feature = GetFeatureStats(stats_view, key)
      stats_dict[key] = {
          'min': feature.num_stats.min,
          'max': feature.num_stats.max,
          'mean': feature.num_stats.mean,
          'median': feature.num_stats.median,
          'std_dev': feature.num_stats.std_dev,
          'max_num_values': feature.num_stats.common_stats.max_num_values
      }

  # Init TFXIO
  tfxio_factory = tfxio_utils.get_tfxio_factory_from_artifact(
      examples=[examples],
      schema=tf_schema,
      telemetry_descriptors=_TELEMETRY_DESCRIPTORS)

  split_and_tfxio = []
  for split in artifact_utils.decode_split_names(examples.split_names):
    split_dir = artifact_utils.get_split_uri([examples], split)
    this_tfxio = tfxio_factory(io_utils.all_files_pattern(split_dir))
    split_and_tfxio.append((split, this_tfxio))

  # Run each split in a Beam pipeline, invoking user code
  # TODO: Replace beam.io.WriteToTFRecord with TFXIO write
  with beam_pipeline:
    for split, tfxio in split_and_tfxio:
      output_split_dir = artifact_utils.get_split_uri([transformed_examples],
                                                      split)
      output_path = os.path.join(output_split_dir, 'data_tfrecord')

      _ = (beam_pipeline
           | 'TFXIORead[{}]'.format(split) >> tfxio.BeamSource()
           | 'Map2Pandas[{}]'.format(split) >>
           beam.Map(lambda record_batch: record_batch.to_pandas())
           | 'Arrow2PandasTypes[{}]'.format(split) >> beam.ParDo(
               Arrow2PandasTypes(), schema=schema_dict)
           | 'UserCode[{}]'.format(split) >> beam.ParDo(
               WrapUserCode, schema=schema_dict, statistics=stats_dict)
           | 'GetExamples[{}]'.format(split) >> beam.ParDo(GetExamples())
           | 'Write[{}]'.format(split) >> beam.io.WriteToTFRecord(
               output_path, file_name_suffix='.gz'))



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Supplement for prima Indians Diabetics example with specifics feature modification.
This module file will be used in the feature selection component example.
"""
from sklearn.feature_selection import \
    SelectKBest as SelectorFunc  # pylint: disable=W0611
from sklearn.feature_selection import chi2

SELECTOR_PARAMS = {"score_func": chi2, "k": 3}
TARGET_FEATURE = 'Outcome'



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tfx_addons.mlmd_client.client."""
import os

from ml_metadata.proto import metadata_store_pb2
from tfx.dsl.component.experimental.annotations import (OutputArtifact,
                                                        Parameter)
from tfx.dsl.component.experimental.decorators import component
from tfx.orchestration.local.local_dag_runner import LocalDagRunner
from tfx.orchestration.pipeline import Pipeline
from tfx.types.standard_artifacts import String

from tfx_addons.mlmd_client import client


@component
def print_component(word: Parameter[str], word_out: OutputArtifact[String]):
  print(word)
  word_out.value = word


def _create_pipeline(root_dir: str):
  comp = print_component(word="test")
  connection_config = metadata_store_pb2.ConnectionConfig()
  connection_config.sqlite.filename_uri = os.path.join(root_dir, "db.sqlite")
  connection_config.sqlite.connection_mode = 3  # READWRITE_OPENCREATE
  return Pipeline(
      pipeline_root=root_dir,
      pipeline_name="client_test",
      metadata_connection_config=connection_config,
      components=[comp],
  )


def test_pipeline_exists(tmpdir):
  pipeline = _create_pipeline(tmpdir.mkdir("test").strpath)
  LocalDagRunner().run(pipeline)
  p = client.MetadataClient.from_pipeline(pipeline)
  assert isinstance(p, client.PipelineContext)


def test_get_artifacts(tmpdir):
  pipeline = _create_pipeline(tmpdir.mkdir("test").strpath)
  LocalDagRunner().run(pipeline)
  p = client.MetadataClient.from_pipeline(pipeline)
  assert isinstance(p, client.PipelineContext)
  assert len(p.get_artifact_by_type_name('String')) == 1



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""" Message Exit Handler component """

import json

from absl import logging
from kfp.pipeline_spec import pipeline_spec_pb2
from tfx import v1 as tfx
from tfx.utils import proto_utils

from tfx_addons.message_exit_handler import constants
from tfx_addons.message_exit_handler.message_providers.base_provider import \
    MessagingType
from tfx_addons.message_exit_handler.message_providers.logging_provider import \
    LoggingMessageProvider
from tfx_addons.message_exit_handler.message_providers.slack_provider import \
    SlackMessageProvider


@tfx.orchestration.experimental.exit_handler
def MessageExitHandler(
    final_status: tfx.dsl.components.Parameter[str],
    on_failure_only: tfx.dsl.components.Parameter[bool] = False,
    message_type: tfx.dsl.components.Parameter[str] = MessagingType.LOGGING.
    value,
    slack_credentials: tfx.dsl.components.Parameter[str] = None,
    decrypt_fn: tfx.dsl.components.Parameter[str] = None,
):
  """
    Exit handler component for TFX pipelines originally developed by
    Digits Financial, Inc.
    The handler notifies the user of the final pipeline status via Slack.

    Args:
        final_status: The final status of the pipeline.
        slack_credentials: (Optional) The credentials to use for the
                           Slack API calls, json format.
        on_failure_only: (Optional) Whether to notify only on failure.
            False is the default.
        message_type: (Optional) The type of message to send.
            Logging is the default.
        decrypt_fn: (Optional) The function to use to decrypt the credentials,
        'tfx_addons.message_exit_handler.component_tests.fake_decryption_fn'

    """

  # parse the final status
  pipeline_task_status = pipeline_spec_pb2.PipelineTaskFinalStatus()
  proto_utils.json_to_proto(final_status, pipeline_task_status)
  logging.debug(f"MessageExitHandler: {final_status}")
  status = json.loads(final_status)

  # leave the exit handler if pipeline succeeded and on_failure_only is True
  if on_failure_only and status["state"] == constants.SUCCESS_STATUS:
    logging.info("MessageExitHandler: Skipping notification on success.")
    return

  # create the message provider
  if message_type == MessagingType.SLACK.value:
    # parse slack credentials
    if not slack_credentials:
      raise ValueError("Slack credentials not provided.")
    provider = SlackMessageProvider(status=status,
                                    credentials=slack_credentials,
                                    decrypt_fn=decrypt_fn)
  elif message_type == MessagingType.LOGGING.value:
    provider = LoggingMessageProvider(status=status)
  else:
    raise ValueError(
        f"MessageExitHandler: Unknown message type: {message_type}")

  provider.send_message()
  message = provider.get_message()
  logging.info(f"MessageExitHandler: {message}")



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""" Message provider interface for logging messages. """

from typing import Dict, Optional

from absl import logging

from tfx_addons.message_exit_handler.message_providers.base_provider import \
    BaseProvider


class LoggingMessageProvider(BaseProvider):
  """Logging message provider."""
  def __init__(
      self,
      status: Dict,
      log_level: Optional[int] = logging.INFO,
  ) -> None:
    super().__init__(status=status)
    self._log_level = log_level

  def send_message(self) -> None:
    logging.log(self._log_level, f"MessageExitHandler: {self._message}")



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Feature Selection component for tfx_addons"""
import importlib
import json
import os

import tensorflow as tf
from tfx.dsl.component.experimental.decorators import component
from tfx.types import artifact, artifact_utils
from tfx.types.standard_artifacts import Examples
from tfx.v1.dsl.components import InputArtifact, OutputArtifact, Parameter
from tfx_bsl.coders import example_coder


# Output artifact containing required data related to feature selection
class FeatureSelectionArtifact(artifact.Artifact):
  """Output artifact containing feature scores from the Feature Selection component"""
  TYPE_NAME = 'Feature Selection'
  PROPERTIES = {
      'scores': artifact.Property(type=artifact.PropertyType.JSON_VALUE),
      'p_values': artifact.Property(type=artifact.PropertyType.JSON_VALUE),
      'selected_features':
      artifact.Property(type=artifact.PropertyType.JSON_VALUE)
  }


# reads and returns data from TFRecords at URI as a list of dictionaries with values as numpy arrays
def _get_data_from_tfrecords(train_uri):
  # get all the data files
  train_uri = [
      os.path.join(train_uri, file_path)
      for file_path in _get_file_list(train_uri)
  ]
  raw_dataset = tf.data.TFRecordDataset(train_uri, compression_type='GZIP')

  np_dataset = []
  for tfrecord in raw_dataset:
    serialized_example = tfrecord.numpy()
    example = example_coder.ExampleToNumpyDict(serialized_example)
    np_dataset.append(example)

  return np_dataset


# returns data in list and nested list formats compatible with sklearn
def _data_preprocessing(np_dataset, target_feature):

  # getting the required data without any metadata
  np_dataset = [{k: v[0]
                 for k, v in example.items()} for example in np_dataset]

  # extracting `y`
  target = [i.pop(target_feature) for i in np_dataset]
  feature_keys = list(np_dataset[0].keys())
  # getting `X`
  input_data = [[i[j] for j in feature_keys] for i in np_dataset]

  return [feature_keys, target, input_data]


# update example with selected features
def _update_example(selected_features, orig_example):
  result = {}
  for key, feature in orig_example.features.feature.items():
    if key in selected_features:
      result[key] = feature

  new_example = tf.train.Example(features=tf.train.Features(feature=result))
  return new_example


# get a list of files for the specified path
def _get_file_list(dir_path):
  file_list = [
      f for f in os.listdir(dir_path)
      if os.path.isfile(os.path.join(dir_path, f))
  ]
  return file_list


@component
def FeatureSelection(  # pylint: disable=C0103
    orig_examples: InputArtifact[Examples],
    feature_selection: OutputArtifact[FeatureSelectionArtifact],
    updated_data: OutputArtifact[Examples],
    module_file: Parameter[str] = None,
    module_path: Parameter[str] = None,
):
  """Runs a user-specified feature selection algorithm on an `Examples` artifact
    Args:
      - orig_examples: An `Examples` input artifact with the data to be
        processed
      - module_file: Python module file containing the configuration
        Example: `modules_files.module_file_a`
        Exactly one of `module_file` and `module_path` should be passed.
        If both are used, module_file would be preferred
      - module_path: Python module path containing the configuration
        Example: `absolute_path/module_files/module_file_a.py` or
        `./module_files/module_file_a.py`
        Exactly one of `module_file` and `module_path` should be passed.
        If both are used, module_file would be preferred

    Module file configuration:
    - SELECTOR_PARAMS: Parameters for SelectorFunc in the form of
      a kwargs dictionary
      Example: {"score_func": chi2, "k": 2}
      Here, `chi2` has been imported from sklearn.feature_selection
    - TARGET_FEATURE: Name of the feature containing target data
    - SelectorFunc: Selector function for univariate feature selection
      Example: SelectKBest, SelectPercentile from sklearn.feature_selection
  """

  # importing the required functions and variables from the module file

  if module_file:
    modules = importlib.import_module(module_file)
  elif module_path:
    module_spec = importlib.util.spec_from_file_location(
        "all_modules", module_path)
    modules = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(modules)

  mod_names = ["SELECTOR_PARAMS", "TARGET_FEATURE", "SelectorFunc"]
  selector_params, target_feature, selector_func = [
      getattr(modules, i) for i in mod_names
  ]

  # uri for the required data
  train_uri = artifact_utils.get_split_uri([orig_examples], 'train')
  np_dataset = _get_data_from_tfrecords(train_uri)
  feature_keys, target_data, input_data = _data_preprocessing(
      np_dataset, target_feature)

  # Select features based on scores
  selector = selector_func(**selector_params)
  selector.fit_transform(input_data, target_data)

  # adding basic info to the updated example artifact as output
  updated_data.split_names = orig_examples.split_names
  updated_data.span = orig_examples.span

  # generate a list of selected features by matching FEATURE_KEYS to selected indices
  selected_features = [
      val for (idx, val) in enumerate(feature_keys)
      if idx in selector.get_support(indices=True)
  ]

  # convert string to array
  split_arr = json.loads(orig_examples.split_names)

  # update feature per split
  for split in split_arr:
    split_uri = artifact_utils.get_split_uri([orig_examples], split)
    new_split_uri = artifact_utils.get_split_uri([updated_data], split)
    os.mkdir(new_split_uri)

    for file in _get_file_list(split_uri):
      split_dataset = tf.data.TFRecordDataset(os.path.join(split_uri, file),
                                              compression_type='GZIP')

      # write the TFRecord
      with tf.io.TFRecordWriter(path=os.path.join(new_split_uri, file),
                                options="GZIP") as writer:
        for split_record in split_dataset.as_numpy_iterator():
          example = tf.train.Example()
          example.ParseFromString(split_record)

          updated_example = _update_example(selected_features, example)
          writer.write(updated_example.SerializeToString())

  # get scores and p-values for artifacts
  selector_scores = selector.scores_
  selector_p_values = selector.pvalues_

  # merge scores and pvalues with feature keys to create a dictionary
  selector_scores_dict = dict(zip(feature_keys, selector_scores))
  selector_pvalues_dict = dict(zip(feature_keys, selector_p_values))

  # populate artifact with the required properties
  feature_selection.scores = selector_scores_dict
  feature_selection.p_values = selector_pvalues_dict
  feature_selection.selected_features = selected_features



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
TFX FeastExampleGen Component Definition

QueryBasedExampleGen
https://github.com/tensorflow/tfx/blob/master/tfx/components/example_gen/component.py

PrestoExampleGen extends QueryBasedExampleGen
https://github.com/tensorflow/tfx/blob/master/tfx/examples/custom_components/presto_example_gen/presto_component/component.py

"""

from typing import List, Optional, Union

import feast
from google.protobuf.struct_pb2 import Struct
from tfx.components.example_gen import component, utils
from tfx.dsl.components.base import executor_spec
from tfx.proto import example_gen_pb2
from tfx.v1.dsl.experimental import RuntimeParameter

from tfx_addons.feast_examplegen import executor


class FeastExampleGen(component.QueryBasedExampleGen):
  """FeastExampleGen is a way for loading offline features. For now we support BQ.

  Example:

    example_gen = FeastExampleGen(
      repo_config=RepoConfig(register="gs://..."),
      entity_query="SELECT user, timestamp from some_user_dataset",
      features=["f1", "f2"],
    )
  """

  EXECUTOR_SPEC = executor_spec.BeamExecutorSpec(executor.Executor)

  def __init__(self,
               repo_config: feast.RepoConfig,
               features: Union[List[str], str, feast.FeatureService],
               entity_query: Optional[str] = None,
               input_config: Union[example_gen_pb2.Input, RuntimeParameter,
                                   None] = None,
               **kwargs):
    """Args:
            repo_config: Feast repo configuration object
            features: List of features to join on the dataset or
              feaure service identifier.
            entity_query: Query used to obtain the entity dataframe.
              Defaults to None.
            **kwargs: kwargs used in QueryBasedExampleGen
    """
    # generate input config
    if bool(entity_query) == bool(input_config):
      raise RuntimeError(
          'Exactly one of query and input_config should be set.')
    input_config = input_config or utils.make_default_input_config(
        entity_query)

    custom_config = Struct()

    # Serialize repo config into a JSON to pass it to the executor
    repo_json = repo_config.json(exclude={"repo_path"}, exclude_unset=True)
    custom_config.update({
        executor._REPO_CONFIG_KEY: repo_json,
    })
    if isinstance(features, str):
      custom_config.update({
          executor._FEATURE_SERVICE_KEY: features,
      })
    elif isinstance(features, list):
      custom_config.update({
          executor._FEATURE_KEY: features,
      })
    elif isinstance(features, feast.FeatureService):
      custom_config.update({
          executor._FEATURE_SERVICE_KEY: features.name,
      })
    else:
      raise RuntimeError(
          f"Features argument not recognized. Found {type(features)} "
          "but expected feature service or list of features")

    # Store configuration as part of a protobuf struct and pack inside custom_config
    custom_config_pbs2 = example_gen_pb2.CustomConfig()
    custom_config_pbs2.custom_config.Pack(custom_config)

    super().__init__(custom_config=custom_config_pbs2,
                     input_config=input_config,
                     **kwargs)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tfx_addons.feature_selection.component"""

import csv
import importlib
import os
from typing import List, Optional, Text

import tensorflow as tf
import tfx
from tfx.orchestration import metadata

from tfx_addons.feature_selection import component


def _get_selected_features(module_file, data_path):
  """Get the correct selected features for testing"""

  data = []

  # importing required configurations
  modules = importlib.import_module(module_file)
  mod_names = ["SELECTOR_PARAMS", "TARGET_FEATURE", "SelectorFunc"]
  selector_params, target_feature, selector_func = [
      getattr(modules, i) for i in mod_names
  ]

  # getting the data
  with open(data_path, 'r') as file:
    my_reader = csv.reader(file, delimiter=',')
    for row in my_reader:
      data.append(row)

  # splitting X (input) and Y (output) from CSV data
  target_idx = data[0].index(target_feature)
  target_data = [i.pop(target_idx) for i in data]

  # runnign the selector function for feature selection
  selector = selector_func(**selector_params)
  selector.fit_transform(data[1:], target_data[1:])

  # getting selected feature names
  selected_indices = selector.get_support(indices=True)
  final_features = set(data[0][idx] for idx in selected_indices)

  return final_features


def _create_pipeline(
    pipeline_name: Text,
    pipeline_root: Text,
    data_root: Text,
    module_path: Text,
    metadata_path: Text,
    beam_pipeline_args: Optional[List[Text]] = None) -> tfx.v1.dsl.Pipeline:
  """Creating sample pipeline with two components: CsvExampleGen and
  FeatureSelection"""

  # specifying the pipeline components
  example_gen = tfx.components.CsvExampleGen(input_base=data_root)
  feature_selection = component.FeatureSelection(
      orig_examples=example_gen.outputs['examples'], module_path=module_path)

  components = [example_gen, feature_selection]

  return tfx.v1.dsl.Pipeline(
      pipeline_name=pipeline_name,
      pipeline_root=pipeline_root,
      components=components,
      metadata_connection_config=metadata.sqlite_metadata_connection_config(
          metadata_path),
      beam_pipeline_args=beam_pipeline_args)


class FeatureSelectionTest(tf.test.TestCase):
  def setUp(self):
    super().setUp()
    self._test_dir = os.path.join(
        os.environ.get('TEST_UNDECLARED_OUTPUTS_DIR', self.get_temp_dir()),
        self._testMethodName)
    self._feature_selection_root = os.path.dirname(__file__)
    self._pipeline_name = 'feature_selection'
    self._data_root = os.path.join(self._feature_selection_root, 'test')
    self._data_path = os.path.join(self._data_root, 'iris.csv')
    self._module_path = os.path.join(self._feature_selection_root, 'example',
                                     'modules', 'iris_module_file.py')
    self._module_file = "tfx_addons.feature_selection.example.modules.\
iris_module_file"

    self._pipeline_root = os.path.join(self._test_dir, 'tfx', 'pipelines',
                                       self._pipeline_name)
    self._metadata_path = os.path.join(self._test_dir, 'tfx', 'metadata',
                                       self._pipeline_name, 'metadata.db')

  def assertExecutedOnce(self, component: Text) -> None:  # pylint: disable=W0621
    """Check the component is executed exactly once."""
    component_path = os.path.join(self._pipeline_root, component)
    self.assertTrue(tfx.dsl.io.fileio.exists(component_path))
    execution_path = os.path.join(component_path, '.system',
                                  'executor_execution')
    execution = tfx.dsl.io.fileio.listdir(execution_path)
    self.assertLen(execution, 1)

  def assertPipelineExecution(self) -> None:
    self.assertExecutedOnce('CsvExampleGen')
    self.assertExecutedOnce('FeatureSelection')

  def testFeatureSelectionPipelineLocal(self):
    tfx.v1.orchestration.LocalDagRunner().run(
        _create_pipeline(pipeline_name=self._pipeline_name,
                         pipeline_root=self._pipeline_root,
                         data_root=self._data_root,
                         module_path=self._module_path,
                         metadata_path=self._metadata_path))

    expected_execution_count = 2  # one each for CsvExampleGen and Feature Selection
    true_selected_features = _get_selected_features(self._module_file,
                                                    self._data_path)

    metadata_config = (
        tfx.orchestration.metadata.sqlite_metadata_connection_config(
            self._metadata_path))
    with metadata.Metadata(metadata_config) as m:
      execution_count = len(m.store.get_executions())
      selected_features_struct = list(
          m.store.get_artifacts_by_type(
              "Feature Selection")[0].properties["selected_features"].
          struct_value.fields.values.__self__["__value__"].list_value.values)
      component_selected_features = set(
          feature.string_value for feature in selected_features_struct)

      # TEST: execution count
      self.assertEqual(expected_execution_count, execution_count)

      # TEST: number of artifacts with TYPE_NAME `Feature Selection`
      self.assertEqual(1,
                       len(m.store.get_artifacts_by_type("Feature Selection")))

      # TEST: number of artifacts with TYPE_NAME `Examples`
      # (one each from CsvExampleGen and FeatureSelection)
      self.assertEqual(2, len(m.store.get_artifacts_by_type("Examples")))

      # TEST: if the features selected by component are correct
      self.assertEqual(component_selected_features, true_selected_features)

    self.assertPipelineExecution()


if __name__ == '__main__':
  tf.compat.v1.enable_v2_behavior()
  tf.test.main()

# _disabled pylint warning `W0621: Redefining name 'component' from outer scope` till an alternate way is found



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Chicago taxi example using TFX schema curation custom component.
base code taken from: https://github.com/tensorflow/tfx/blob/master/tfx/examples/custom_components/hello_world/example/taxi_pipeline_hello.py

This example demonstrate the use of schema curation custom component.
user defined function `schema_fn` defined in `module_file.py` is used
to change feature `tips` from required to optional.

"""

import os
import tempfile
import urllib
from typing import Text

import absl
import tfx
from tfx.components import CsvExampleGen, SchemaGen, StatisticsGen
from tfx.orchestration import metadata, pipeline
from tfx.orchestration.local import local_dag_runner

from tfx_addons.schema_curation.component import component

# downloading data and setting up required paths
_data_root = tempfile.mkdtemp(prefix='tfx-data')
DATA_PATH = 'https://raw.githubusercontent.com/tensorflow/tfx/master/tfx/examples/chicago_taxi_pipeline/data/simple/data.csv'
_data_filepath = os.path.join(_data_root, "data.csv")
urllib.request.urlretrieve(DATA_PATH, _data_filepath)

_pipeline_name = 'taxi_pipeline'
_tfx_root = tfx.__path__[0]
_pipeline_root = os.path.join(_tfx_root, 'pipelines', _pipeline_name)
_metadata_path = os.path.join(_tfx_root, 'metadata', _pipeline_name,
                              'metadata.db')


def _create_pipeline(pipeline_name: Text, pipeline_root: Text, data_root: Text,
                     metadata_path: Text) -> pipeline.Pipeline:
  """Implements the chicago taxi pipeline with TFX."""

  # Brings data into the pipeline or otherwise joins/converts training data.
  example_gen = CsvExampleGen(input_base=data_root)

  # Computes statistics over data for visualization and example validation.
  statistics_gen = StatisticsGen(examples=example_gen.outputs['examples'])

  # inferes a schema
  schema_gen = SchemaGen(statistics=statistics_gen.outputs['statistics'],
                         infer_feature_shape=True)

  # modifies infered schema with use of udf `schema_fn` defined in module file
  schema_curation = component.SchemaCuration(
      schema=schema_gen.outputs['schema'],
      module_file=os.path.join('schemacomponent', 'example', 'module_file.py'))

  return pipeline.Pipeline(
      pipeline_name=pipeline_name,
      pipeline_root=pipeline_root,
      components=[example_gen, statistics_gen, schema_gen, schema_curation],
      enable_cache=True,
      metadata_connection_config=metadata.sqlite_metadata_connection_config(
          metadata_path))


# To run this pipeline from the python CLI:
#   $python taxi_pipeline_hello.py
if __name__ == '__main__':
  absl.logging.set_verbosity(absl.logging.INFO)
  local_dag_runner.LocalDagRunner().run(
      _create_pipeline(pipeline_name=_pipeline_name,
                       pipeline_root=_pipeline_root,
                       data_root=_data_root,
                       metadata_path=_metadata_path))



""" Null preprocessing, for minimal testing """
from absl import logging


# pylint: disable=unused-argument
def preprocessing_fn(df, schema, statistics):
  logging.info('Running null preprocessing')
  return df



# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tfx_addons.pandas_transform.component."""

import apache_beam as beam
import tensorflow as tf
from packaging.version import Version, parse
from tfx import v1 as tfx
from tfx.orchestration import data_types
from tfx.types import (artifact_utils, channel_utils, standard_artifacts,
                       standard_component_specs)

from tfx_addons.pandas_transform import component


class ComponentTest(tf.test.TestCase):
  def setUp(self):
    super().setUp()
    examples_artifact = standard_artifacts.Examples()
    examples_artifact.split_names = artifact_utils.encode_split_names(
        ['train', 'eval'])
    self.examples = channel_utils.as_channel([examples_artifact])
    self.schema = channel_utils.as_channel([standard_artifacts.Schema()])
    statistics_artifact = standard_artifacts.ExampleStatistics()
    statistics_artifact.split_names = artifact_utils.encode_split_names(
        ['train', 'eval'])
    self.statistics = channel_utils.as_channel([statistics_artifact])

  def _verify_outputs(self, pandas_transform):
    self.assertEqual(
        standard_artifacts.Examples.TYPE_NAME, pandas_transform.outputs[
            standard_component_specs.TRANSFORMED_EXAMPLES_KEY].type_name)

  def test_construct_from_module_file(self):
    module_file = './null_preprocessing.py'
    if parse(tfx.__version__) >= Version('1.8.0'):
      pandas_transform = component.PandasTransform(
          examples=self.examples,
          schema=self.schema,
          statistics=self.statistics,
          module_file=module_file,
          beam_pipeline=beam.Pipeline())
    else:
      pandas_transform = component.PandasTransform(examples=self.examples,
                                                   schema=self.schema,
                                                   statistics=self.statistics,
                                                   module_file=module_file)
    self._verify_outputs(pandas_transform)
    self.assertEqual(
        module_file, pandas_transform.exec_properties[
            standard_component_specs.MODULE_FILE_KEY])

  def test_construct_with_parameter(self):
    module_file = data_types.RuntimeParameter(name='module-file', ptype=str)
    if parse(tfx.__version__) >= Version('1.8.0'):
      pandas_transform = component.PandasTransform(
          examples=self.examples,
          schema=self.schema,
          statistics=self.statistics,
          module_file=module_file,
          beam_pipeline=beam.Pipeline())
    else:
      pandas_transform = component.PandasTransform(examples=self.examples,
                                                   schema=self.schema,
                                                   statistics=self.statistics,
                                                   module_file=module_file)
    self._verify_outputs(pandas_transform)
    self.assertJsonEqual(
        str(module_file),
        str(pandas_transform.exec_properties[
            standard_component_specs.MODULE_FILE_KEY]))


#   NOTE: This test is not currently working because
#         Python-function components don't currently
#         propagate exceptions correctly.  b/238368874
#
#   def test_construct_missing_user_module(self):
#     module_file = './missing_file.py'
#     with self.assertRaises(ImportError):
#         pandas_transform = component.PandasTransform(
#           examples=self.examples,
#           schema=self.schema,
#           statistics=self.statistics,
#           module_file=module_file
#       )

if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Firebase Publisher TFX Component.

The FirebasePublisher is used to deploy model to Firebase ML.
"""

from typing import Dict, List, Optional

from tfx import types
from tfx.dsl.components.base import base_component, executor_spec
from tfx.types import standard_artifacts
from tfx.types.component_spec import ChannelParameter, ExecutionParameter

from tfx_addons.firebase_publisher import executor

MODEL_KEY = "model"
PUSHED_MODEL_KEY = "pushed_model"
MODEL_BLESSING_KEY = "model_blessing"


class FirebasePublisherSpec(types.ComponentSpec):
  """ComponentSpec for TFX FirebasePublisher Component."""

  PARAMETERS = {
      "display_name": ExecutionParameter(type=str),
      "storage_bucket": ExecutionParameter(type=str),
      "app_name": ExecutionParameter(type=str, optional=True),
      "tags": ExecutionParameter(type=List, optional=True),
      "options": ExecutionParameter(type=Dict, optional=True),
      "credential_path": ExecutionParameter(type=str, optional=True),
  }
  INPUTS = {
      MODEL_KEY:
      ChannelParameter(type=standard_artifacts.Model, optional=True),
      MODEL_BLESSING_KEY:
      ChannelParameter(type=standard_artifacts.ModelBlessing, optional=True),
  }
  OUTPUTS = {
      PUSHED_MODEL_KEY: ChannelParameter(type=standard_artifacts.PushedModel),
  }


class FirebasePublisher(base_component.BaseComponent):
  """Component for pushing model to Firebase ML.

  The `FirebasePublisher` is a [TFX
  Component](https://www.tensorflow.org/tfx/guide/understanding_tfx_pipelines#component)
  that pushes a model to Firebase ML.
  This component publishes a model from an upstream component
  such as [`Trainer`](https://www.tensorflow.org/tfx/guide/trainer) to
  Firebase ML that hosts ML models that mobile applications can consume.
  As such, it provides better interoperability between ML models and mobile
  applications that need to consume them. To find more about Firebase ML,
  please refer to the official
  [webpage](https://firebase.google.com/products/ml).
  """

  SPEC_CLASS = FirebasePublisherSpec
  EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(executor.Executor)

  def __init__(
      self,
      display_name: str,
      storage_bucket: str,
      app_name: Optional[str] = None,
      tags: Optional[List[str]] = None,
      options: Optional[Dict] = None,
      credential_path: Optional[str] = None,
      model: Optional[types.BaseChannel] = None,
      model_blessing: Optional[types.BaseChannel] = None,
  ):
    """The FirebasePublisher TFX component.
       FirebasePublisher deploys a trained or blessed model to Firebase ML.
       This is designed to work as a downstream component of Trainer and
       Evaluator(optional) components. Trainer gives trained model, and
       Evaluator gives information whether the trained model is blessed or
       not after evaluation of the model. This component only publishes a
       model when it is blessed. If Evaluator is not specified, the input
       model will always be published.
       **Important Note:** Firebase ML allows to host a model under 40mb.
       if the model size exceeds 40mb, RuntimError will be raised.


    Args:
      display_name: name to identify a hosted model in Firebase ML. this should
        be a unique value because it will be used to search a existing model to
        update.
      storage_bucket: GCS bucket where the hosted model will be stored.
      app_name: the name of Firebase app to determine the scope. if not given,
        default app name, '[DEFAULT]' will be used.
      tags: tags to be attached to the hosted ML model. any meaningful tags will
        be helpful to understand your model in Firebase ML platform.
      options: additional configurations to be passed to initialize Firebase
        app. refer to the official document to find out which options are
        available at [`initialize_app()`]
        (https://firebase.google.com/docs/reference/admin/python/firebase_admin#initialize_app).
      credential_path: location of GCS or local file system where the Service
        Account(SA) Key file is. the SA should have sufficeint permissions to
        create and view hosted models in Firebase ML. If this parameter is not
        given, Application Default Credentials will be used in GCP environment.
      model: a TFX input channel containing a Model artifact. this is usually
        comes from the standard [`Trainer`]
        (https://www.tensorflow.org/tfx/guide/trainer) component.
      model_blessing: a TFX input channel containing a ModelBlessing artifact.
        this is usually comes from the standard [`Evaluator`]
        (https://www.tensorflow.org/tfx/guide/evaluator) component.
    Returns:
      a TFX output channel containing a PushedModel artifact. It contains
      information where the model is published at and whether the model is
      pushed or not.
    Raises:
      RuntimeError: when model size exceeds 40mb.
    Example:
      Basic usage example:
      ```py
      trainer = Trainer(...)
      evaluator = Evaluator(...)

      fb_publisher = FirebasePublisher(
        display_name="model_on_firebase",
        storage_bucket="firebase_ml", # only the bucket name without gs://
        credential_path="gs://abc.json",
        model=trainer.outputs["model"],
        model_blessing=evaluator.outputs["blessing"]
      )
      ```
    """

    pushed_model = types.Channel(type=standard_artifacts.PushedModel)

    spec = FirebasePublisherSpec(app_name=app_name,
                                 display_name=display_name,
                                 tags=tags,
                                 storage_bucket=storage_bucket,
                                 options=options,
                                 credential_path=credential_path,
                                 model=model,
                                 model_blessing=model_blessing,
                                 pushed_model=pushed_model)

    super().__init__(spec=spec)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
Tests for feast_component.executor.
"""

from __future__ import absolute_import, division, print_function

import datetime
from typing import Optional
from unittest import mock

import apache_beam as beam
import pytest
from apache_beam.testing import util
from google.cloud import bigquery
from google.protobuf.struct_pb2 import Struct

try:
  import feast
  from feast.infra.offline_stores.bigquery import (BigQueryOfflineStoreConfig,
                                                   BigQueryRetrievalJob)

except ImportError:
  pytest.skip("feast not available, skipping", allow_module_level=True)

import tensorflow as tf
from tfx.extensions.google_cloud_big_query import utils
from tfx.proto import example_gen_pb2
from tfx.utils import proto_utils

from tfx_addons.feast_examplegen import executor


class FooRetrievalJob(BigQueryRetrievalJob):
  def to_bigquery(
      self,
      job_config: bigquery.QueryJobConfig = None,  # pylint: disable=unused-argument
      timeout: int = 1800,  # pylint: disable=unused-argument
      retry_cadence: int = 10,  # pylint: disable=unused-argument
  ) -> Optional[str]:  # pylint: disable=invalid-name, unused-argument
    return _MockReadFromFeast()


@beam.ptransform_fn
def _MockReadFromFeast(pipeline, query):  # pylint: disable=invalid-name
  del query  # Unused arg
  mock_query_results = [{
      'timestamp': datetime.datetime.utcfromtimestamp(4.2e6),
      'i': 1,
      'i2': [2, 3],
      'b': True,
      'f': 2.0,
      'f2': [2.7, 3.8],
      's': 'abc',
      's2': ['abc', 'def']
  }]
  return pipeline | beam.Create(mock_query_results)


def _mock_custom_config():
  _REPO_CONFIG_KEY = "repo_conf"  # pylint: disable=invalid-name
  _FEATURE_KEY = "feature_refs"  # pylint: disable=invalid-name

  offline_store = BigQueryOfflineStoreConfig()
  repo_config = feast.RepoConfig(provider='gcp',
                                 project='default',
                                 offline_store=offline_store)
  repo_conf = repo_config.json(exclude={"repo_path"}, exclude_unset=True)
  feature_refs = ['feature1', 'feature2']
  config_struct = Struct()
  config_struct.update({
      _REPO_CONFIG_KEY: repo_conf,
      _FEATURE_KEY: feature_refs
  })
  custom_config_pbs2 = example_gen_pb2.CustomConfig()
  custom_config_pbs2.custom_config.Pack(config_struct)
  custom_config = proto_utils.proto_to_json(custom_config_pbs2)
  return custom_config


def _mock_load_custom_config(custom_config):  # pylint: disable=unused-argument
  offline_store = BigQueryOfflineStoreConfig()
  repo_config = feast.RepoConfig(provider='gcp',
                                 project='default',
                                 offline_store=offline_store)
  repo_conf = repo_config.json(exclude={"repo_path"}, exclude_unset=True)
  feature_refs = ['feature1', 'feature2']

  return {
      executor._REPO_CONFIG_KEY: repo_conf,  # pylint: disable=protected-access
      executor._FEATURE_KEY: feature_refs  # pylint: disable=protected-access
  }


def _mock_get_retrieval_job(entity_query, custom_config):  # pylint: disable=invalid-name, unused-argument
  del entity_query, custom_config  # pylint: disable=unused-argument

  return FooRetrievalJob("SELECT * FROM  `test.test`", "Client",
                         bigquery.QueryJobConfig(dry_run=True), "names")


def _mock_get_historical_features(exec_properties):  # pylint: disable=invalid-name, unused-argument
  del exec_properties  # Unused arg
  return BigQueryRetrievalJob("SELECT * FROM `test.test`", "Client", None,
                              "names")


class ExecutorTest(tf.test.TestCase):
  def setUp(self):
    # Mock BigQuery result schema.
    self._schema = [
        bigquery.SchemaField("timestamp", 'TIMESTAMP', mode='REQUIRED'),
        bigquery.SchemaField('i', 'INTEGER', mode='REQUIRED'),
        bigquery.SchemaField('i2', 'INTEGER', mode='REPEATED'),
        bigquery.SchemaField('b', 'BOOLEAN', mode='REQUIRED'),
        bigquery.SchemaField('f', 'FLOAT', mode='REQUIRED'),
        bigquery.SchemaField('f2', 'FLOAT', mode='REPEATED'),
        bigquery.SchemaField('s', 'STRING', mode='REQUIRED'),
        bigquery.SchemaField('s2', 'STRING', mode='REPEATED'),
    ]
    super().setUp()

  def testLoadCustomConfig(self):
    offline_store = BigQueryOfflineStoreConfig()
    repo_config = feast.RepoConfig(provider='gcp',
                                   project='default',
                                   offline_store=offline_store)
    repo_conf = repo_config.json(exclude={"repo_path"}, exclude_unset=True)
    feature_refs = ['feature1', 'feature2']
    config_struct = Struct()
    config_struct.update({
        executor._REPO_CONFIG_KEY: repo_conf,  # pylint: disable=protected-access
        executor._FEATURE_KEY: feature_refs  # pylint: disable=protected-access
    })
    custom_config_pbs2 = example_gen_pb2.CustomConfig()
    custom_config_pbs2.custom_config.Pack(config_struct)
    custom_config = proto_utils.proto_to_json(custom_config_pbs2)
    deseralized_conn = executor._load_custom_config(custom_config)  # pylint: disable=protected-access
    truth_config = _mock_load_custom_config(" ")
    self.assertEqual(deseralized_conn, truth_config)

  @mock.patch.multiple(
      executor,
      _load_custom_config=_mock_load_custom_config,
  )
  def testLoadFeastFeatureStore(self):
    custom_config = executor._load_custom_config(" ")  # pylint: disable=protected-access
    feature_store = executor._load_feast_feature_store(custom_config)  # pylint: disable=protected-access
    self.assertEqual(feature_store.config.project, 'default')  # pylint: disable=protected-access

  @mock.patch.multiple(executor,
                       _load_custom_config=_mock_load_custom_config,
                       _get_retrieval_job=_mock_get_retrieval_job)
  @mock.patch.multiple(
      utils,
      ReadFromBigQuery=_MockReadFromFeast,
  )
  @mock.patch.object(bigquery, 'Client')
  def testFeastToExample(self, mock_client):
    mock_client.return_value.query.return_value.result.return_value.schema = \
      self._schema
    with beam.Pipeline() as pipeline:
      examples = (
          pipeline
          | 'ToTFExample' >> executor._FeastToExampleTransform(  # pylint: disable=protected-access
              exec_properties={
                  '_beam_pipeline_args': [],
                  'custom_config': _mock_custom_config()
              },
              split_pattern='SELECT i, f, s FROM `fake`'))
      feature = {}
      feature['timestamp'] = tf.train.Feature(float_list=tf.train.FloatList(
          value=[4200000]))
      feature['i'] = tf.train.Feature(int64_list=tf.train.Int64List(value=[1]))
      feature['i2'] = tf.train.Feature(int64_list=tf.train.Int64List(
          value=[2, 3]))
      feature['b'] = tf.train.Feature(int64_list=tf.train.Int64List(value=[1]))
      feature['f'] = tf.train.Feature(float_list=tf.train.FloatList(
          value=[2.0]))
      feature['f2'] = tf.train.Feature(float_list=tf.train.FloatList(
          value=[2.7, 3.8]))
      feature['s'] = tf.train.Feature(bytes_list=tf.train.BytesList(
          value=[tf.compat.as_bytes('abc')]))
      feature['s2'] = tf.train.Feature(bytes_list=tf.train.BytesList(
          value=[tf.compat.as_bytes('abc'),
                 tf.compat.as_bytes('def')]))
      example_proto = tf.train.Example(features=tf.train.Features(
          feature=feature)).SerializeToString(deterministic=True)

      util.assert_that(examples, util.equal_to([example_proto]))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for the custom xgboost Evaluator module."""

import os

import apache_beam as beam
import pandas as pd
import tensorflow as tf
import tensorflow_model_analysis as tfma
import xgboost as xgb
from apache_beam.testing import util
from google.protobuf import text_format
from tensorflow_metadata.proto.v0 import schema_pb2
from tfx.types import channel_utils, standard_artifacts
from tfx_bsl.tfxio import tensor_adapter, test_util

from tfx_addons.xgboost_evaluator import component, xgboost_predict_extractor


class XGBoostPredictExtractorTest(
    tfma.test.testutil.TensorflowModelAnalysisTest):
  def setUp(self):
    """Function that sets up a schema, some examples, and some metadata
    for use in the tests."""

    super().setUp()
    self._eval_export_dir = os.path.join(self._getTempDir(), 'eval_export')
    self._create_xgboost_model(self._eval_export_dir)
    self._eval_config = tfma.EvalConfig(
        model_specs=[tfma.ModelSpec(name='', label_key="label")])
    self._eval_shared_model = (
        xgboost_predict_extractor.custom_eval_shared_model(
            eval_saved_model_path=self._eval_export_dir,
            model_name='',
            eval_config=self._eval_config))
    self._schema = text_format.Parse(
        """
        feature {
          name: "age"
          type: FLOAT
        }
        feature {
          name: "language"
          type: FLOAT
        }
        feature {
          name: "label"
          type: INT
        }
        """, schema_pb2.Schema())
    self._tfx_io = test_util.InMemoryTFExampleRecord(
        schema=self._schema, raw_record_column_name=tfma.ARROW_INPUT_COLUMN)
    self._tensor_adapter_config = tensor_adapter.TensorAdapterConfig(
        arrow_schema=self._tfx_io.ArrowSchema(),
        tensor_representations=self._tfx_io.TensorRepresentations())
    self._examples = [
        self._makeExample(age=3.0, language=1.0, label=1),
        self._makeExample(age=3.0, language=0.0, label=0),
        self._makeExample(age=4.0, language=1.0, label=1),
        self._makeExample(age=5.0, language=0.0, label=0),
    ]

  def testMakeXGBoostPredictExtractor(self):
    """Tests that predictions are made from extracts for a single model."""
    feature_extractor = tfma.extractors.FeaturesExtractor(self._eval_config)
    prediction_extractor = (
        xgboost_predict_extractor.make_xgboost_predict_extractor(
            self._eval_shared_model, self._eval_config))

    def check_result(item):
      try:
        # Regular assert used due to errors with pytest when using self.AssertEqual
        assert item['labels'].shape == item['predictions'].shape

      except AssertionError as err:
        raise util.BeamAssertException(err)

    with beam.Pipeline() as pipeline:
      _ = (
          pipeline
          | 'Create' >> beam.Create(
              [e.SerializeToString() for e in self._examples])
          | 'BatchExamples' >> self._tfx_io.BeamSource()
          | 'InputsToExtracts' >> tfma.BatchedInputsToExtracts()  # pylint: disable=no-value-for-parameter
          | feature_extractor.stage_name >> feature_extractor.ptransform
          | prediction_extractor.stage_name >> prediction_extractor.ptransform
          | beam.Map(check_result))  # the test

  def testMakeXGBoostPredictExtractorWithMultiModels(self):
    """Tests that predictions are made from extracts for multiple models."""
    eval_config = tfma.EvalConfig(model_specs=[
        tfma.ModelSpec(name='model1', label_key="label"),
        tfma.ModelSpec(name='model2', label_key="label"),
    ])
    eval_export_dir_1 = os.path.join(self._eval_export_dir, '1')
    self._create_xgboost_model(eval_export_dir_1)
    eval_shared_model_1 = xgboost_predict_extractor.custom_eval_shared_model(
        eval_saved_model_path=eval_export_dir_1,
        model_name='model1',
        eval_config=eval_config)
    eval_export_dir_2 = os.path.join(self._eval_export_dir, '2')
    self._create_xgboost_model(eval_export_dir_2)
    eval_shared_model_2 = xgboost_predict_extractor.custom_eval_shared_model(
        eval_saved_model_path=eval_export_dir_2,
        model_name='model2',
        eval_config=eval_config)

    feature_extractor = tfma.extractors.FeaturesExtractor(self._eval_config)
    prediction_extractor = (
        xgboost_predict_extractor.make_xgboost_predict_extractor(
            eval_shared_model={
                'model1': eval_shared_model_1,
                'model2': eval_shared_model_2,
            },
            eval_config=eval_config))

    def check_result(item):
      try:
        # Regular assert used due to errors with pytest when using self.AssertEqual
        assert len(item['labels']) == len(item['predictions'])
        assert 'model1' in item['predictions'][0]
        assert 'model2' in item['predictions'][0]

      except AssertionError as err:
        raise util.BeamAssertException(err)

    with beam.Pipeline() as pipeline:
      _ = (
          pipeline
          | 'Create' >> beam.Create(
              [e.SerializeToString() for e in self._examples])
          | 'BatchExamples' >> self._tfx_io.BeamSource()
          | 'InputsToExtracts' >> tfma.BatchedInputsToExtracts()  # pylint: disable=no-value-for-parameter
          | feature_extractor.stage_name >> feature_extractor.ptransform
          | prediction_extractor.stage_name >> prediction_extractor.ptransform
          | beam.Map(check_result))  # the test

  def test_custom_eval_shared_model(self):
    """Tests that an EvalSharedModel is created with a custom xgboost loader."""
    model_file = os.path.basename(self._eval_shared_model.model_path)
    self.assertEqual(model_file, 'model.json')
    model = self._eval_shared_model.model_loader.construct_fn()
    self.assertIsInstance(model, xgb.Booster)

  def test_custom_extractors(self):
    """Tests that the xgboost extractor is used when creating extracts."""
    extractors = xgboost_predict_extractor.custom_extractors(
        self._eval_shared_model, self._eval_config,
        self._tensor_adapter_config)
    self.assertLen(extractors, 6)
    self.assertIn('XGBoostPredict',
                  [extractor.stage_name for extractor in extractors])

  def test_component(self):
    examples = standard_artifacts.Examples()
    model_exports = standard_artifacts.Model()
    evaluator = component.XGBoostEvaluator(
        examples=channel_utils.as_channel([examples]),
        model=channel_utils.as_channel([model_exports]),
        example_splits=['eval'])

    module_file = xgboost_predict_extractor.get_module_file()
    self.assertEqual(standard_artifacts.ModelEvaluation.TYPE_NAME,
                     evaluator.outputs['evaluation'].type_name)
    self.assertEqual(module_file, evaluator.exec_properties["module_file"])

  def _create_xgboost_model(self, eval_export_dir):
    """Creates and pickles a toy xgboost model.

    Args:
        eval_export_dir: Directory to store a pickled xgboost model. This
            directory is created if it does not exist.
    """
    train = pd.DataFrame({"age": [3, 0], "language": [4, 1]})
    label = pd.DataFrame({"label": [0, 1]})
    param = {'max_depth': 2, 'eta': 1, 'objective': 'binary:logistic'}
    train = xgb.DMatrix(train, label=label)
    model = xgb.train(param, train)

    os.makedirs(eval_export_dir)
    model.save_model(os.path.join(eval_export_dir, "model.json"))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Firebase Publisher runner module.

This module handles the workflow to publish
machine learning model to Firebase ML.
"""
import glob
import os
import tempfile
from typing import Any, Dict, List, Tuple

import firebase_admin
import tensorflow as tf
from absl import logging
from firebase_admin import credentials, ml
from firebase_admin.ml import ListModelsPage, TFLiteModelSource
from tfx.dsl.io import fileio
from tfx.utils import io_utils

_SIZE_LIMIT_MB = 80


def prepare_fb_download_model(app_name: str, credential_path: str,
                              storage_bucket: str, model_path: str,
                              options: Dict[str, Any]) -> str:
  """Initializes a Firebase app, and download the
     target model to a temporary directory."""
  tmp_dir = tempfile.gettempdir()
  credential = None

  if credential_path is not None:
    tmp_credential_path = os.path.join(tmp_dir, "credentials.json")
    io_utils.copy_file(credential_path, tmp_credential_path)
    credential = credentials.Certificate(tmp_credential_path)
    logging.info(
        "credentials are copied into a temporary directory in local filesystem"
    )

  options["storageBucket"] = storage_bucket

  firebase_admin.initialize_app(credential=credential,
                                options=options,
                                name=app_name)
  logging.info("firebase app initialization is completed")

  tmp_model_path = os.path.join(tmp_dir, "model")
  io_utils.copy_dir(model_path, tmp_model_path)

  return tmp_model_path


def get_model_path_and_type(tmp_model_path) -> Tuple[bool, str]:
  """Fetches model path and flag if the model is TFLite."""
  tflite_files = glob.glob(f"{tmp_model_path}/**/*.tflite")
  is_tflite = len(tflite_files) > 0
  model_path = tflite_files[0] if is_tflite else tmp_model_path

  return is_tflite, model_path


def upload_model_to_gcs(is_tflite: bool, model_path: str) -> TFLiteModelSource:
  """upload model to GCS. If SavedModel, it will be
     converted to TFLite under the hood"""
  if is_tflite:
    source = ml.TFLiteGCSModelSource.from_tflite_model_file(model_path)
  else:
    source = ml.TFLiteGCSModelSource.from_saved_model(model_path)

  return source


def check_model_size(source: TFLiteModelSource):
  """Checks model size and if it exceeds 40 MB (Firebase limit)
     a RuntimeError is raised."""
  gcs_path_for_uploaded_file = source.as_dict().get('gcsTfliteUri')
  with tf.io.gfile.GFile(gcs_path_for_uploaded_file) as f:
    file_size_in_mb = f.size() / (1 << 20)

  if file_size_in_mb > _SIZE_LIMIT_MB:
    fileio.remove(gcs_path_for_uploaded_file)
    raise RuntimeError(
        f"the file size {file_size_in_mb} exceeds"
        f"the limit of {_SIZE_LIMIT_MB}. Uploaded file is removed.")


def is_model_present(model_list: ListModelsPage):
  return len(model_list.models) > 0


def update_model(model_list: ListModelsPage, source: TFLiteModelSource,
                 tags: List[str], model_version: str):
  """update existing model in Firebase ML with the latest version"""
  tags.append(model_version)

  # get the first match model. if there are multiple models
  # with the same name, only the first match will be selected.
  # so, the model should be named unique
  model = model_list.models[0]

  model.tags = tags
  model.model_format = ml.TFLiteFormat(model_source=source)

  updated_model = ml.update_model(model)
  ml.publish_model(updated_model.model_id)

  logging.info("model exists, so it is updated")


def create_model(display_name: str, source: TFLiteModelSource, tags: List[str],
                 model_version: str):
  """create a new model"""
  tags.append(model_version)

  tflite_format = ml.TFLiteFormat(model_source=source)
  model = ml.Model(
      display_name=display_name,
      tags=tags,
      model_format=tflite_format,
  )

  # Add the model to your Firebase project and publish it
  new_model = ml.create_model(model)
  ml.publish_model(new_model.model_id)

  logging.info(
      "TFLite model didn't exist, so it will be first created and pushed.")


def deploy_model_for_firebase_ml(
    app_name: str,
    display_name: str,
    storage_bucket: str,
    tags: List[Any],
    options: Dict[str, Any],
    model_path: str,
    model_version: str,
    credential_path: str,
) -> str:
  """Executes ML model deployment workflow to Firebase ML.
       Refer to the FirebasePublisher component in component.py
       for generic description of each parameter. This docstring
       only explains how the workflow works.

       step #1.
       workflow begins by initializing firebase app with the
       given app_name, storage_bucket, options, and credentials.
       currently, Firebase ML only let us host/upload ML model
       from the local filesystem, so it downloads a model from
       the upstream components to a temporary directory.

       step #2.
       search the list of hosted models with display_name. this
       has to be done before uploading/hosting any ML models.
       otherwise, the number of hosted models is greater than 0,
       hence it will try to update the existing model even though
       it has to create a new model.

       step #3.
       get the model path of the downloaded model in the local
       filesystem. along the way, it searches for any files whose
       extension is '*.tflite'. when it finds tflite model file,
       is_tflite flag is marked and the file path is obtained.
       Otherwise (in case of SavedModel), the directory of the
       SavedModel is obtained.

       step #4.
       upload/host the ML model to the designated gcs_storage bucket.
       along the way, if the given model is SavedModel, it will be
       directly converted to TFLite format before uploading

       step #5.
       check the size of the uploaded TFLite model. currently,
       Firebase ML allows us to host a model whose size is under 40mb.
       if the given TFLite model's size exceeds 40mb, then the
       uploaded model files will be deleted (rollback), and
       RuntimeError exceptioin will be raised.

       step #6.
       update the existing model. the existing tags will be replaced.
       while keeping the given tags are reserved, newly granted
       model_version will be appended to the list.

       step #7.
       create a new model.

    Returns:
        str: the GCS storage bucket path where the actual model is hosted
    """
  # step 1
  tmp_model_path = prepare_fb_download_model(app_name, credential_path,
                                             storage_bucket, model_path,
                                             options)

  # step 2
  model_list = ml.list_models(list_filter=f"display_name={display_name}")

  # step 3
  is_tflite, model_path = get_model_path_and_type(tmp_model_path)

  # step 4
  source = upload_model_to_gcs(is_tflite, model_path)

  # step 5
  check_model_size(source)

  if is_model_present(model_list.models):
    # step 6
    update_model(model_list, source, tags, model_version)
  else:
    # step 7
    create_model(display_name, source, tags, model_version)

  # step 8
  return source.as_dict().get('gcsTfliteUri')



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Firebase Publisher TFX Component Executor.

The Firebase Publisher Executor calls the workflow handler
runner.deploy_model_for_firebase_ml().
"""

import time
from typing import Any, Dict, List

from tfx import types
from tfx.components.pusher import executor as tfx_pusher_executor
from tfx.types import artifact_utils, standard_component_specs

from tfx_addons.firebase_publisher import runner

_APP_NAME_KEY = "app_name"
_DISPLAY_NAME_KEY = "display_name"
_STORAGE_BUCKET_KEY = "storage_bucket"
_TAGS_KEY = "tags"
_OPTIONS_KEY = "options"
_CREDENTIAL_PATH_KEY = "credential_path"


class Executor(tfx_pusher_executor.Executor):
  """Pushes a model to Firebase ML."""
  def Do(
      self,
      input_dict: Dict[str, List[types.Artifact]],
      output_dict: Dict[str, List[types.Artifact]],
      exec_properties: Dict[str, Any],
  ):
    """Overrides the tfx_pusher_executor to leverage some of utility methods

    Args:
      input_dict: Input dict from input key to a list of artifacts, including:
        - model_export: a TFX input channel containing a Model artifact.
        - model_blessing: a TFX input channel containing a ModelBlessing
          artifact.
      output_dict: Output dict from key to a list of artifacts, including:
        - pushed_model: a TFX output channel containing a PushedModel artifact.
          It contains information where the model is published at and whether
          the model is pushed or not.
      exec_properties: An optional dict of execution properties, including:
        - display_name: name to identify a hosted model in Firebase ML.
            this should be a unique value because it will be used to search
            a existing model to update.
        - storage_bucket: GCS bucket where the hosted model will be stored.
        - app_name: the name of Firebase app to determine the scope.
        - tags: tags to be attached to the hosted ML model.
        - credential_path: location of GCS or local file system where the
          Service Account(SA) Key file is.
        - options: additional configurations to be passed to initialize Firebase
          app.

    Raises:
      RuntimeError: when the size of model exceeds 40mb.
    """
    self._log_startup(input_dict, output_dict, exec_properties)

    model_push = artifact_utils.get_single_instance(
        output_dict[standard_component_specs.PUSHED_MODEL_KEY])
    if not self.CheckBlessing(input_dict):
      self._MarkNotPushed(model_push)
      return
    model_path = self.GetModelPath(input_dict)
    model_version_name = f"v{int(time.time())}"

    pushed_model_path = runner.deploy_model_for_firebase_ml(
        app_name=exec_properties.get(_APP_NAME_KEY, '[DEFAULT]'),
        display_name=exec_properties.get(_DISPLAY_NAME_KEY),
        storage_bucket=exec_properties.get(_STORAGE_BUCKET_KEY),
        credential_path=exec_properties.get(_CREDENTIAL_PATH_KEY, None),
        tags=exec_properties.get(_TAGS_KEY, []),
        options=exec_properties.get(_OPTIONS_KEY, {}),
        model_path=model_path,
        model_version=model_version_name,
    )

    self._MarkPushed(model_push, pushed_destination=pushed_model_path)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""" Message provider interface for slack messages. """

from typing import Dict, Optional

from absl import logging
from pydantic import BaseModel
from slack import WebClient
from slack.errors import SlackApiError
from tfx.utils import import_utils

from tfx_addons.message_exit_handler.message_providers.base_provider import \
    BaseProvider


class SlackCredentials(BaseModel):
  """Pydantic class to de/serialize the slack credentials."""
  slack_token: str
  slack_channel_id: str


class SlackMessageProvider(BaseProvider):
  """Slack message provider."""
  def __init__(self,
               status: Dict,
               credentials: str,
               decrypt_fn: Optional[str] = None) -> None:
    super().__init__(status=status)

    if not credentials:
      raise ValueError("Slack credentials not provided.")

    credentials = SlackCredentials.parse_raw(credentials)
    self._slack_channel_id = credentials.slack_channel_id
    self._slack_token = credentials.slack_token

    if decrypt_fn:
      module_path, fn_name = decrypt_fn.rsplit(".", 1)
      logging.info(
          f"MessageExitHandler: Importing {fn_name} from {module_path} "
          "to decrypt credentials.")
      fn = import_utils.import_func_from_module(module_path, fn_name)
      self._slack_channel_id = fn(self._slack_channel_id)
      self._slack_token = fn(self._slack_token)

    self._client = WebClient(token=self._slack_token)

  def send_message(self) -> None:
    try:
      response = self._client.chat_postMessage(channel=self._slack_channel_id,
                                               text=self._message)
      logging.info(f"MessageExitHandler: Slack response: {response}")
    except SlackApiError as e:
      logging.error(
          f"MessageExitHandler: Slack API error: {e.response['error']}")



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Init module for TFX."""

import importlib as _importlib

from .version import _PKG_METADATA, __version__

_ACTIVE_MODULES = [
    "__version__",
] + list(_PKG_METADATA.keys())


def __getattr__(name):  # pylint: disable=C0103
  # PEP-562: Lazy loaded attributes on python modules
  # NB(gcasassaez): We lazy load to avoid issues with dependencies not installed
  # for some subpackes
  if name in _ACTIVE_MODULES:
    return _importlib.import_module("." + name, __name__)
  raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""XGBoost Evaluator component."""

from tfx import v1 as tfx

from tfx_addons.xgboost_evaluator import xgboost_predict_extractor


class XGBoostEvaluator(tfx.components.Evaluator):
  """A custom Evaluator component made for XGBoost. Keeps everything the same,
  except inputs the custom module file containing the XGBoost Extractor."""
  def __init__(self, **kwargs):
    if 'module_file' in kwargs:
      raise ValueError('XGBoostEvaluator does not accept custom module_file')
    super().__init__(module_file=xgboost_predict_extractor.get_module_file(),
                     **kwargs)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Executor test for the sampling component's executor."""

import filecmp
import os
import tempfile
from unittest.mock import MagicMock, patch

import apache_beam as beam
import tensorflow as tf
from absl.testing import absltest
from apache_beam.testing.util import assert_that, equal_to
from tfx.components.util import tfxio_utils
from tfx.dsl.io import fileio
from tfx.types import artifact_utils, standard_artifacts
from tfx.utils import io_utils, json_utils

from tfx_addons.sampling import executor, spec


class ExecutorTest(absltest.TestCase):
  def _validate_output(self, output, splits, num=1):
    def generate_elements(data):
      for i in range(len(data[list(data.keys())[0]])):
        gen = dict()
        for key in data.keys():
          if data[key][i] and len(data[key][i]) > 0:
            gen[key] = data[key][i][0]
          else:
            gen[key] = ""
        yield gen

    tfxio_factory = tfxio_utils.get_tfxio_factory_from_artifact(
        examples=[output], telemetry_descriptors=[])
    for split in splits:
      tfxio = tfxio_factory(
          io_utils.all_files_pattern(
              artifact_utils.get_split_uri([output], split)))

      with beam.Pipeline() as p:
        data = (p
                | 'TFXIORead[%s]' % split >> tfxio.BeamSource()
                | 'DictConversion' >> beam.Map(lambda x: x.to_pydict())
                | 'ConversionCleanup' >> beam.FlatMap(generate_elements)
                | 'MapToLabel' >> beam.Map(lambda x: (x['label'], x))
                | 'CountPerKey' >> beam.combiners.Count.PerKey()
                | 'FilterNull' >> beam.Filter(lambda x: x[0])
                | 'Values' >> beam.Values()
                | 'Distinct' >> beam.Distinct()
                | 'Count' >> beam.combiners.Count.Globally())

    assert_that(data, equal_to([num]))

  def _validate_same(self, dir0, dir1):
    comp = filecmp.dircmp(dir0, dir1)
    self.assertTrue(comp.left_only == [])
    self.assertTrue(comp.right_only == [])
    self.assertTrue(comp.diff_files == [])

  def _run_exec(self, exec_properties):
    source_data_dir = os.path.join(os.path.dirname(__file__), 'data')
    output_data_dir = os.path.join(
        os.environ.get('TEST_UNDECLARED_OUTPUTS_DIR', tempfile.mkdtemp()),
        self._testMethodName)
    fileio.makedirs(output_data_dir)

    examples = standard_artifacts.Examples()
    examples.uri = os.path.join(source_data_dir, "example_gen")
    examples.split_names = artifact_utils.encode_split_names(['train', 'eval'])

    input_dict = {
        spec.SAMPLER_INPUT_KEY: [examples],
    }

    # Create output dict.
    output = standard_artifacts.Examples()
    output.uri = output_data_dir
    output_dict = {
        spec.SAMPLER_OUTPUT_KEY: [output],
    }

    # Run executor.
    under = executor.Executor()
    under.Do(input_dict, output_dict, exec_properties)

    return output

  def testDo(self):
    exec_properties = {
        spec.SAMPLER_LABEL_KEY:
        'label',
        spec.SAMPLER_SPLIT_KEY:
        json_utils.dumps(['train']),
        # List needs to be serialized before being passed into Do function.
        spec.SAMPLER_COPY_KEY:
        True,
        spec.SAMPLER_SHARDS_KEY:
        1,
        spec.SAMPLER_CLASSES_KEY:
        json_utils.dumps([]),
        spec.SAMPLER_SAMPLE_KEY:
        True,
    }

    output = self._run_exec(exec_properties)

    # Check outputs.
    self._validate_output(output, ['train'])
    self.assertTrue(fileio.exists(os.path.join(output.uri, 'Split-train')))
    self.assertTrue(fileio.exists(os.path.join(output.uri, 'Split-eval')))

  def testKeepClasses(self):
    exec_properties = {
        spec.SAMPLER_LABEL_KEY:
        'label',
        spec.SAMPLER_SPLIT_KEY:
        json_utils.dumps(['train']),
        # List needs to be serialized before being passed into Do function.
        spec.SAMPLER_COPY_KEY:
        True,
        spec.SAMPLER_SHARDS_KEY:
        1,
        spec.SAMPLER_CLASSES_KEY:
        json_utils.dumps(['None']),
        spec.SAMPLER_SAMPLE_KEY:
        spec.SamplingStrategy.UNDERSAMPLE,
    }

    output = self._run_exec(exec_properties)

    # Check outputs.
    self._validate_output(output, ['train'], 2)

  def testShards(self):
    exec_properties = {
        spec.SAMPLER_LABEL_KEY: 'label',
        spec.SAMPLER_SPLIT_KEY: json_utils.dumps(
            ['train']
        ),  # List needs to be serialized before being passed into Do function.
        spec.SAMPLER_COPY_KEY: True,
        spec.SAMPLER_SHARDS_KEY: 20,
        spec.SAMPLER_CLASSES_KEY: json_utils.dumps([]),
        spec.SAMPLER_SAMPLE_KEY: spec.SamplingStrategy.UNDERSAMPLE,
    }

    output = self._run_exec(exec_properties)

    # Check outputs.
    out = os.path.join(output.uri, 'Split-train')
    self.assertTrue(
        len([
            name for name in os.listdir(out)
            if os.path.isfile(os.path.join(out, name))
        ]) == 20)

  def testSplits(self):
    exec_properties = {
        spec.SAMPLER_LABEL_KEY: 'label',
        spec.SAMPLER_SPLIT_KEY: json_utils.dumps(
            ['train', 'eval']
        ),  # List needs to be serialized before being passed into Do function.
        spec.SAMPLER_COPY_KEY: True,
        spec.SAMPLER_SHARDS_KEY: 1,
        spec.SAMPLER_CLASSES_KEY: json_utils.dumps([]),
        spec.SAMPLER_SAMPLE_KEY: spec.SamplingStrategy.UNDERSAMPLE,
    }

    output = self._run_exec(exec_properties)

    # Check outputs.
    self._validate_output(output, ['train'])
    self._validate_output(output, ['eval'])

  def testCopy(self):
    exec_properties = {
        spec.SAMPLER_LABEL_KEY: 'label',
        spec.SAMPLER_SPLIT_KEY: json_utils.dumps(
            ['train']
        ),  # List needs to be serialized before being passed into Do function.
        spec.SAMPLER_COPY_KEY: False,
        spec.SAMPLER_SHARDS_KEY: 1,
        spec.SAMPLER_CLASSES_KEY: json_utils.dumps([]),
        spec.SAMPLER_SAMPLE_KEY: spec.SamplingStrategy.UNDERSAMPLE,
    }

    output = self._run_exec(exec_properties)

    self.assertFalse(fileio.exists(os.path.join(output.uri, 'Split-eval')))

  # Pipeline tests below!

  def testFilter(self):
    assert executor.filter_null([5, 5])  # return
    assert executor.filter_null([0, 0])  # return
    assert not executor.filter_null(["", ""])  # no return
    assert not executor.filter_null(["", 5])  # no return
    assert not executor.filter_null([5, 5], keep_null=True)  # no return
    assert not executor.filter_null([0, 0], keep_null=True)  # no return
    assert executor.filter_null(["", ""], keep_null=True)  # return
    assert not executor.filter_null([5, 5], null_vals=["5"])  # no return
    assert executor.filter_null([5, 5], keep_null=True,
                                null_vals=["5"])  # return
    assert executor.filter_null(["", ""], keep_null=True,
                                null_vals=["5"])  # return

  @patch("random.random", MagicMock(side_effect=[0.2, 0.6, 0.7] * 10))
  def testPipelineMin(self):
    dataset = [("1", 1), ("1", 1), ("1", 1), ("2", 2), ("2", 2), ("2", 2),
               ("2", 2), ("3", 3), ("3", 3), ("", 0)]
    expected = [1, 1, 2, 2, 3, 3, 0]

    with beam.Pipeline() as p:
      data = p | beam.Create(dataset)
      merged = executor.sample_examples(data, None,
                                        spec.SamplingStrategy.UNDERSAMPLE)
      assert_that(merged, equal_to(expected))

  @patch("random.random", MagicMock(side_effect=[0.2, 0.6, 0.7] * 10))
  def testPipelineMax(self):
    dataset = [("1", 1), ("1", 1), ("1", 1), ("2", 2), ("2", 2), ("2", 2),
               ("2", 2), ("3", 3), ("3", 3), ("", 0)]
    expected = [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 0]

    with beam.Pipeline() as p:
      data = p | beam.Create(dataset)
      merged = executor.sample_examples(data, None,
                                        spec.SamplingStrategy.OVERSAMPLE)
      assert_that(merged, equal_to(expected))

  def testMinimum(self):
    dataset = [("1", 1), ("1", 1), ("1", 1), ("2", 2), ("2", 2), ("2", 2),
               ("2", 2), ("3", 3), ("3", 3), ("", 0)]

    def find_minimum(elements):
      return min(elements or [0])

    with beam.Pipeline() as p:
      val = (p
             | beam.Create(dataset)
             | "CountPerKey" >> beam.combiners.Count.PerKey()
             | "FilterNullCount" >>
             beam.Filter(lambda x: executor.filter_null(x, null_vals=None))
             | "Values" >> beam.Values()
             | "GetSample" >> beam.CombineGlobally(find_minimum))
      assert_that(val, equal_to([2]))

  def testMaximum(self):
    dataset = [("1", 1), ("1", 1), ("1", 1), ("2", 2), ("2", 2), ("2", 2),
               ("2", 2), ("3", 3), ("3", 3), ("", 0)]

    def find_maximum(elements):
      return max(elements or [0])

    with beam.Pipeline() as p:
      val = (p
             | beam.Create(dataset)
             | "CountPerKey" >> beam.combiners.Count.PerKey()
             | "FilterNullCount" >>
             beam.Filter(lambda x: executor.filter_null(x, null_vals=None))
             | "Values" >> beam.Values()
             | "GetSample" >> beam.CombineGlobally(find_maximum))
      assert_that(val, equal_to([4]))

  def testException(self):
    exec_properties = {
        spec.SAMPLER_LABEL_KEY: 'label',
        spec.SAMPLER_SPLIT_KEY: json_utils.dumps(
            ['bad']
        ),  # List needs to be serialized before being passed into Do function.
        spec.SAMPLER_COPY_KEY: False,
        spec.SAMPLER_SHARDS_KEY: 1,
        spec.SAMPLER_CLASSES_KEY: json_utils.dumps([]),
        spec.SAMPLER_SAMPLE_KEY: spec.SamplingStrategy.UNDERSAMPLE,
    }

    with self.assertRaises(ValueError) as context:
      self._run_exec(exec_properties)
    self.assertTrue('Invalid split name bad is not in input artifact!' in str(
        context.exception))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Message Providers supported by the Message Exit Handler component.

Currently supported:
* Logging
* Slack

"""

import enum
from typing import Dict, Text

from tfx_addons.message_exit_handler import constants


class MessagingType(enum.Enum):
  """Determines the type of message to send."""

  LOGGING = "logging"
  SLACK = "slack"


class BaseProvider:
  """Message provider interface."""
  def __init__(self, status: Dict) -> None:
    self._status = status
    self._message = self.set_message(status)

  @staticmethod
  def set_message(status) -> str:
    """Set the message to be sent."""
    # parse the Vertex paths
    # structure: projects/{project}/locations/{location}/pipelineJobs/{pipeline_job}
    elements = status["pipelineJobResourceName"].split("/")
    project = elements[1]
    location = elements[3]
    job_id = elements[-1]

    # Generate message
    if status["state"] == constants.SUCCESS_STATUS:
      message = (
          ":tada: "
          f"Pipeline job *{job_id}* ({project}) completed successfully.\n")
    else:
      message = f":scream: Pipeline job *{job_id}* ({project}) failed."
      message += f"\n>{status['error']['message']}"

    message += f"\nhttps://console.cloud.google.com/vertex-ai/locations/{location}/pipelines/runs/{job_id}"
    return message

  def get_message(self) -> Text:
    """Get the message to be sent."""
    return self._message



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tfx_addons.firebase_publisher.runner."""

from unittest import mock
from unittest.mock import Mock

import tensorflow as tf

from tfx_addons.firebase_publisher import runner


class RunnerTest(tf.test.TestCase):
  def testModelExistancy(self):
    model_list = Mock()
    model_list.models = ['model1']
    self.assertTrue(runner.is_model_present(model_list))

    model_list.models = []
    self.assertFalse(runner.is_model_present(model_list))

  @mock.patch('tfx_addons.firebase_publisher.runner.glob.glob')
  def testModelPathAndType(self, mock_glob):
    tmp_model_path = "/tmp/saved_model"

    mock_glob.return_value = [f"{tmp_model_path}/model.tflite"]
    is_tflite, model_path = runner.get_model_path_and_type(tmp_model_path)
    self.assertTrue(is_tflite)
    self.assertEquals(f"{tmp_model_path}/model.tflite", model_path)

    mock_glob.return_value = []
    is_tflite, model_path = runner.get_model_path_and_type(tmp_model_path)
    self.assertFalse(is_tflite)
    self.assertEquals(tmp_model_path, model_path)

  @mock.patch('tfx_addons.firebase_publisher.runner.fileio')
  @mock.patch('tfx_addons.firebase_publisher.runner.tf.io.gfile.GFile')
  def testCheckModelSize(self, mock_gfile, mock_fileio):
    mock_source = Mock()
    mock_source.as_dict.get.return_value = "mock_return"

    mock_gfile().__enter__.return_value.size.return_value = 83886080
    mock_gfile().__exit__ = Mock(return_value=False)

    try:
      runner.check_model_size(mock_source)
    except RuntimeError:
      self.fail("Runtime error occured unexpectedly")

    mock_fileio.remove()
    mock_gfile().__enter__.return_value.size.return_value = 83886081
    mock_gfile().__exit__ = Mock(return_value=False)
    with self.assertRaises(RuntimeError):
      runner.check_model_size(mock_source)


if __name__ == "__main__":
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Predict extractor for xgboost models."""

import copy
import os
from typing import Dict, Iterable, List, Text

import apache_beam as beam
import numpy as np
import pandas as pd
import tensorflow_model_analysis as tfma

try:
  # Path for TFMA >= 0.34
  from tensorflow_model_analysis.utils import (
      DoFnWithModels, verify_and_update_eval_shared_models)
except ImportError:
  # Path for TFMA < 0.34
  from tensorflow_model_analysis import DoFnWithModels, verify_and_update_eval_shared_models

import xgboost as xgb
from tfx_bsl.tfxio import tensor_adapter

_PREDICT_EXTRACTOR_STAGE_NAME = 'XGBoostPredict'


def make_xgboost_predict_extractor(
    eval_shared_model: tfma.EvalSharedModel,
    eval_config: tfma.EvalConfig,
) -> tfma.extractors.Extractor:
  """Creates an extractor for performing predictions using a xgboost model.

  The extractor's PTransform loads and runs the serving pickle against
  every extract yielding a copy of the incoming extracts with an additional
  extract added for the predictions keyed by tfma.PREDICTIONS_KEY. The model
  inputs are searched for under tfma.FEATURES_KEY.

  Args:
    eval_shared_model: Shared model (single-model evaluation).

  Returns:
    Extractor for extracting predictions.
  """
  eval_shared_models = verify_and_update_eval_shared_models(eval_shared_model)
  return tfma.extractors.Extractor(
      stage_name=_PREDICT_EXTRACTOR_STAGE_NAME,
      ptransform=_ExtractPredictions(  # pylint: disable=no-value-for-parameter
          eval_shared_models={m.model_name: m
                              for m in eval_shared_models},
          eval_config=eval_config))


@beam.typehints.with_input_types(tfma.Extracts)
@beam.typehints.with_output_types(tfma.Extracts)
class _TFMAPredictionDoFn(DoFnWithModels):
  """A DoFn that loads the models and predicts."""
  def __init__(self, eval_shared_models: Dict[Text, tfma.EvalSharedModel],
               eval_config: tfma.EvalConfig):
    super().__init__(
        {k: v.model_loader
         for k, v in eval_shared_models.items()})
    self._eval_config = eval_config

  def setup(self):
    """Function that accesses and saves the chosen feature keys and label key."""

    # models are loaded and stored into self._loaded_models below
    super().setup()
    self._feature_keys = None
    self._label_key = None

    # check to see whether each of these loaded models has a corresponding model spec
    if self._eval_config:
      label_config = self.extract_model_specs()
      for name in self._loaded_models:
        if name not in label_config:
          raise ValueError(f"Missing model spec for loaded model {name}.")

    for name, loaded_model in self._loaded_models.items():
      feature_keys = loaded_model.feature_names
      if self._feature_keys and self._label_key:
        assert self._feature_keys == feature_keys, (
            f'Features mismatch in loaded models. Expected {self._feature_keys}'
            f', got {feature_keys} instead.')
        assert self._label_key == label_config[name], (
            f'Label mismatch in loaded models. Expected "{self._label_key}"'
            f', got "{label_config[name]}" instead.')
      elif feature_keys and label_config[name]:
        self._feature_keys = feature_keys
        self._label_key = label_config[name]
      else:
        raise ValueError(
            f'Missing feature or label keys in loaded model {name}.')

  def extract_model_specs(self):
    label_specs = {}
    for config in self._eval_config.model_specs:
      label_specs[config.name] = config.label_key
    return label_specs

  def process(self, elem: tfma.Extracts) -> Iterable[tfma.Extracts]:
    """Uses loaded models to make predictions on batches of data.

    Args:
      elem: An extract containing batched features.

    Yields:
      Copy of the original extracts with predictions added for each model. If
      there are multiple models, a list of dicts keyed on model names will be
      added, with each value corresponding to a prediction for a single sample.
    """
    # Build feature and label vectors because xgboost cannot read tf.Examples.
    features = []
    labels = []
    result = copy.copy(elem)
    features_key = result[tfma.FEATURES_KEY]
    # ToDo(gcasassaez): Remove fork once we have TFMA stable release.
    if isinstance(features_key, list):
      for features_dict in features_key:
        features_row = [features_dict[key] for key in self._feature_keys]
        features.append(np.concatenate(features_row))
        labels.append(features_dict[self._label_key])
      result[tfma.LABELS_KEY] = np.concatenate(labels)
    else:
      for key in self._feature_keys:
        for i, v in enumerate(features_key[key]):
          if i >= len(features):
            features.append([])
          features[i].append(v)
      result[tfma.LABELS_KEY] = features_key[self._label_key]

    features = xgb.DMatrix(pd.DataFrame(features, columns=self._feature_keys))

    # Generate predictions for each model.
    for model_name, loaded_model in self._loaded_models.items():
      preds = loaded_model.predict(features)
      if len(self._loaded_models) == 1:
        result[tfma.PREDICTIONS_KEY] = preds
      elif tfma.PREDICTIONS_KEY not in result:
        result[tfma.PREDICTIONS_KEY] = [{model_name: pred} for pred in preds]
      else:
        for i, pred in enumerate(preds):
          result[tfma.PREDICTIONS_KEY][i][model_name] = pred
    yield result


@beam.ptransform_fn
@beam.typehints.with_input_types(tfma.Extracts)
@beam.typehints.with_output_types(tfma.Extracts)
def _ExtractPredictions(  # pylint: disable=invalid-name
    extracts: beam.pvalue.PCollection,
    eval_shared_models: Dict[Text, tfma.EvalSharedModel],
    eval_config: tfma.EvalConfig,
) -> beam.pvalue.PCollection:
  """A PTransform that adds predictions and possibly other tensors to extracts.

  Args:
    extracts: PCollection of extracts with inputs keyed by tfma.INPUTS_KEY.
    eval_shared_models: Shared model parameters keyed by model name.

  Returns:
    PCollection of Extracts updated with the predictions.
  """
  return extracts | 'Predict' >> beam.ParDo(
      _TFMAPredictionDoFn(eval_shared_models, eval_config))


def _custom_model_loader_fn(model_path: Text):
  """Returns a function that loads a xgboost model."""
  def loader(path):
    model = xgb.Booster()
    model.load_model(path)
    return model

  return lambda: loader(model_path)


# TFX Evaluator will call the following functions.
def custom_eval_shared_model(eval_saved_model_path, model_name, eval_config,
                             **kwargs) -> tfma.EvalSharedModel:
  """Returns a single custom EvalSharedModel."""
  model_path = os.path.join(eval_saved_model_path, 'model.json')
  return tfma.default_eval_shared_model(
      eval_saved_model_path=model_path,
      model_name=model_name,
      eval_config=eval_config,
      custom_model_loader=tfma.ModelLoader(
          construct_fn=_custom_model_loader_fn(model_path)),
      add_metrics_callbacks=kwargs.get('add_metrics_callbacks'))


def custom_extractors(
    eval_shared_model: tfma.MaybeMultipleEvalSharedModels,
    eval_config: tfma.EvalConfig,
    tensor_adapter_config: tensor_adapter.TensorAdapterConfig,
) -> List[tfma.extractors.Extractor]:
  """Returns default extractors plus a custom prediction extractor."""
  predict_extractor = make_xgboost_predict_extractor(eval_shared_model,
                                                     eval_config)
  return tfma.default_extractors(eval_shared_model=eval_shared_model,
                                 eval_config=eval_config,
                                 tensor_adapter_config=tensor_adapter_config,
                                 custom_predict_extractor=predict_extractor)


def get_module_file():
  return os.path.abspath(__file__)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Supplement for chicago taxi pipeline example with specifics schema modification.
This module file will be used in the custom schema curation component.
"""

import tensorflow_data_validation as tfdv

# TFX schema curation component will call this function.


def schema_fn(schema):
  """modifies the infered schema.
  Args:
    schema:schema generated by SchemaGen component of tfx
  """
  #changing "tips" into optional feature
  feature = tfdv.get_feature(schema, 'tips')
  feature.presence.min_fraction = 0.9

  return schema



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for Slack Provider functions."""

from unittest.mock import patch

import tensorflow as tf

from tfx_addons.message_exit_handler import constants
from tfx_addons.message_exit_handler.message_providers import slack_provider


class SlackProviderTest(tf.test.TestCase):
  @staticmethod
  def get_final_status(state: str = constants.SUCCESS_STATUS,
                       error: str = "") -> str:
    """Assemble final status for tests"""
    status = {
        "state":
        state,
        "error":
        error,
        "pipelineJobResourceName":
        ("projects/test-project/locations/"
         "us-central1/pipelineJobs/test-pipeline-job"),
    }
    if error:
      status.update({"error": {"message": error}})
    return status

  @patch(
      'tfx_addons.message_exit_handler.message_providers.slack_provider.WebClient'
  )
  def test_slack_message_provider(self, web_client_mock):
    final_status = self.get_final_status()
    credentials = slack_provider.SlackCredentials(
        slack_token="test-token", slack_channel_id="test-channel").json()

    message_provider = slack_provider.SlackMessageProvider(
        final_status, credentials)
    message_provider.send_message()
    web_client_mock.assert_called_once()
    web_client_mock.assert_called_with(token='test-token')

  @patch(
      'tfx_addons.message_exit_handler.message_providers.slack_provider.WebClient'
  )
  def test_slack_message_provider_with_decrypt_fn(self, mock_web_client):
    final_status = self.get_final_status()
    credentials = slack_provider.SlackCredentials(
        slack_token="test-token", slack_channel_id="test-channel").json()

    message_provider = slack_provider.SlackMessageProvider(
        final_status,
        credentials,
        decrypt_fn=
        'tfx_addons.message_exit_handler.component_test.fake_decryption_fn')
    message_provider.send_message()
    mock_web_client.assert_called_once()
    mock_web_client.assert_called_with(token='TEST-TOKEN')


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
Tests for tfx_addons.feast_examplegen.component.
"""

import pytest

try:
  import feast
except ImportError:
  pytest.skip("feast not available, skipping", allow_module_level=True)

from tfx.v1.proto import Input

from tfx_addons.feast_examplegen.component import FeastExampleGen


def test_init_valid():
  entity_query = 'SELECT user FROM fake_db'
  repo_config = feast.RepoConfig(provider='local', project='default')
  FeastExampleGen(repo_config=repo_config,
                  features=['feature1', 'feature2'],
                  entity_query='SELECT user FROM fake_db')
  FeastExampleGen(repo_config=repo_config,
                  features='feature_service1',
                  entity_query='SELECT user FROM fake_db')
  FeastExampleGen(repo_config=repo_config,
                  features=['feature1', 'feature2'],
                  input_config=Input(splits=[
                      Input.Split(name='train', pattern=entity_query),
                      Input.Split(name='eval', pattern=entity_query),
                  ]))


def test_input_and_entity():
  entity_query = 'SELECT user FROM fake_db'
  repo_config = feast.RepoConfig(provider='local', project='default')
  with pytest.raises(RuntimeError):

    FeastExampleGen(repo_config=repo_config,
                    features=['feature1', 'feature2'],
                    entity_query=entity_query,
                    input_config=Input(splits=[
                        Input.Split(name='train', pattern=entity_query),
                        Input.Split(name='eval', pattern=entity_query),
                    ]))
